import enum
from _typeshed import Incomplete

PY_37: Incomplete
PY_38: Incomplete
PY_39: Incomplete
PY_310: Incomplete
PY_311_MIN: Incomplete
PY_310_MIN: Incomplete
PY_39_MIN: Incomplete
PY_38_MIN: Incomplete
PY_37_MIN: Incomplete
PEP_585_COMPATIBLE = PY_39_MIN
PEP_586_COMPATIBLE = PY_38_MIN

class Sentinel(enum.Enum):
    MISSING: Incomplete
