from typing import Generator, Optional

class CodeLines:
    def __init__(self) -> None: ...
    def append(self, line: str) -> None: ...
    def indent(self, expr: Optional[str] = ...) -> Generator[None, None, None]: ...
    def as_text(self) -> str: ...
    def reset(self) -> None: ...
